# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ironic.drivers import ipmi
from ironic.drivers.modules.intel_ipmi import management


class IntelIPMIHardware(ipmi.IPMIHardware):
    """Intel IPMI hardware type.

    Uses ``ipmitool`` to implement power and management.
    Provides serial console implementations via ``shellinabox`` or ``socat``.
    Supports Intel SST-PP feature.
    """
    @property
    def supported_management_interfaces(self):
        """List of supported management interfaces."""
        return [management.IntelIPMIManagement]
